#!/usr/bin/env python3

import numpy as np
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.animation as animation


glider = np.array([[0,1,0],
                   [0,0,1],
                   [1,1,1]],
                  dtype=bool)


c10orthogonal = np.array([[0,1,1,0,0,1,1,0],
                          [0,0,0,1,1,0,0,0],
                          [0,0,0,1,1,0,0,0],
                          [1,0,1,0,0,1,0,1],
                          [1,0,0,0,0,0,0,1],
                          [0,0,0,0,0,0,0,0],
                          [1,0,0,0,0,0,0,1],
                          [0,1,1,0,0,1,1,0],
                          [0,0,1,1,1,1,0,0],
                          [0,0,0,0,0,0,0,0],
                          [0,0,0,1,1,0,0,0],
                          [0,0,0,1,1,0,0,0]],
                         dtype=bool)


def gamegrid(w, h, entities):
    """Creates the game grid on which the Game of Life is to be executed.

    Args:
        w (int): Width of the grid.
        h (int): Height of the grid.
        entities (List[Tuple[np.ndarray, int, int]]):
            List of start entities together with their initial positions that
            are to be placed at the specified position on the grid. You may
            safely assume, that the positional entries in the list do not
            break the boundaries of the game grid.

    Returns:
        np.ndarray: The game grid modelled as an (h, w) numpy-array of type
        ``bool``.
    """
    grid = np.zeros((h,w), dtype=bool)
    for (entity, x, y) in entities:
        add_entity(grid, entity, x, y)
    return grid


def add_entity(grid, entity, y, x):
    """Adds an ``entity`` to the given ``grid`` at the specified position.

    Args:
        grid (np.ndarray): The original game grid.
        entity (np.ndarray): The entity that should be added.
        y (int): The y-position the entity shall be added at.
        x (int): The x-position the entity shall be added at.

    Returns:
        np.ndarray: The updated game grid with the entity starting at position
        (y, x).
    """
    # TODO
    height, width = np.shape(entity)
    grid[y:y+height, x:x+width] = entity
    return grid

'''
def next_step_2(grid):
    n, m = np.shape(grid)
    for i in range(n):
        for j in range(m):
            if i != 0 and j != m - 1 and i != n - 1 and j != 0 and grid[i, j] == 0:
                if grid[i - 1, j] == 0 and grid[i + 1, j] == 1 and grid[i, j - 1] == 1 and grid[i, j + 1] == 1:
                    add_entity(grid, np.ones(1, 1), i, j)
                elif grid[i - 1, j] == 1 and grid[i + 1, j] == 0 and grid[i, j - 1] == 1 and grid[i, j + 1] == 1:
                    add_entity(grid, np.ones(1, 1), i, j)
                elif grid[i - 1, j] == 1 and grid[i + 1, j] == 1 and grid[i, j - 1] == 0 and grid[i, j + 1] == 1:
                    add_entity(grid, np.ones(1, 1), i, j)
                elif grid[i - 1, j] == 1 and grid[i + 1, j] == 1 and grid[i, j - 1] == 1 and grid[i, j + 1] == 0:
                    add_entity(grid, np.ones(1, 1), i, j)
            else:
                if i == 0 and (j != 0 or j != m - 1):
                    if grid[i, j] == 0 and grid[i, j - 1] == 1 and grid[i, j + 1] == 1:
                        add_entity(grid, np.ones(1, 1), i, j)
    return grid
'''

def next_step(grid):
    """Updates the game grid ``grid`` according to the game rules.

    Args:
        grid (np.ndarray): The game grid.

    Returns:
        np.ndarray: The game grid after one time step. You can read the
        rules according to which you should update each cell in your
        exercise sheet.
    """
    # TODO
    height, width = grid.shape
    next_grid = np.zeros_like(grid)
    for i in range(height):
        for j in range(width):
            # Count the number of alive neighbors
            num_alive = (
                int(grid[(i-1)%height, (j-1)%width]) +
                int(grid[(i-1)%height, j]) +
                int(grid[(i-1)%height, (j+1)%width]) +
                int(grid[i, (j-1)%width]) +
                int(grid[i, (j+1)%width]) +
                int(grid[(i+1)%height, (j-1)%width]) +
                int(grid[(i+1)%height, j]) +
                int(grid[(i+1)%height, (j+1)%width])
            )
            # Apply the game rules
            if grid[i, j] == 0 and num_alive == 3:
                next_grid[i, j] = 1
            elif grid[i, j] == 1 and num_alive < 2:
                next_grid[i, j] = 0
            elif grid[i, j] == 1 and (num_alive == 2 or num_alive == 3):
                next_grid[i, j] = grid[i, j]
            else:
                next_grid[i, j] = 0
    return next_grid


def gameoflife(grid):
    """Animates the Game of Life using Matplotlib.

    Args:
        grid (np.ndarray): The game grid with which the animation starts.
    """
    fig, ax = plt.subplots()
    mat = ax.matshow(grid, cmap=cm.gray_r)
    ani = animation.FuncAnimation(fig, lambda _: mat.set_data(next_step(grid)),
                                  frames=100,
                                  interval=50,
                                  blit=False)
    plt.show()


def main():
    """The Main-Function of the programme. Is executed whenever this file is
    executed at top level.

    Hier kann beliebiger Testcode stehen, der bei der Korrektur vollständig
    ignoriert wird.
    """
    grid = gamegrid(40, 40, [(glider, 13, 4),
                             (c10orthogonal, 25, 25)])
    gameoflife(grid)

if __name__ == "__main__": main()